// icbv.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "videoInput.h"

using namespace std;

int g_select_object = 0;
int g_range_selected = 0;
CvPoint g_origin;
CvRect g_selection;
CvPoint g_selection_bottom;
IplImage *g_image = 0;

bool GetLaserPos(IplImage *g_image, double &pic_pos_x, double &pic_pos_y)
{
	pic_pos_x = 0;
	pic_pos_y = 0;
	int ii = 0;
	int point_count = 0;
	int delta = g_image->widthStep - g_selection.width*3;
	unsigned char *src_pic = (unsigned char *)g_image->imageData;
	const int treshold = 225 * 3;
//	int loc_x = 0, loc_y = 0;

	ii = g_selection.y * g_image->widthStep + g_selection.x * 3;
	for (int y = g_selection.y; y < g_selection_bottom.y; ++y, ii += delta)
	{
		//std::cout << y << std::endl;
		for (int x = g_selection.x; x < g_selection_bottom.x; ++x, ii += 3)
		{
			
			unsigned int ir, ig, ib;
			
			ir = src_pic[ii];
			ig = src_pic[ii + 1];
			ib = src_pic[ii + 2];
			if (ir + ib + ig > treshold)
			{
//				std::cout << "found pixel! " << x << ", " << y << std::endl;
//				std::cout << ir << ", " << ig << ", " << ib << std::endl;
				pic_pos_x += (double)x;
				pic_pos_y += (double)y;
				++point_count;
			}
		}
		//++loc_y;
	}
//	std::cout << point_count << std::endl;
	if (point_count < 4)
	{
		return false;
	}
	
	pic_pos_x /= point_count;
	pic_pos_y /= point_count;

	//std::cout << pic_pos_x << ", " << pic_pos_y << std::endl;
	return true;

}

void on_mouse( int event, int x, int y, int flags, void* param )
{
    if( !g_image )
        return;

    if( g_image->origin )
        y = g_image->height - y;

    if( g_select_object )
    {
        g_selection.x = MIN(x, g_origin.x);
        g_selection.y = MIN(y, g_origin.y);
        g_selection.width = g_selection.x + CV_IABS(x - g_origin.x);
        g_selection.height = g_selection.y + CV_IABS(y - g_origin.y);
        
        g_selection.x = MAX( g_selection.x, 0 );
        g_selection.y = MAX( g_selection.y, 0 );
        g_selection.width = MIN( g_selection.width, g_image->width );
        g_selection.height = MIN( g_selection.height, g_image->height );
        g_selection.width -= g_selection.x;
        g_selection.height -= g_selection.y;

		g_selection_bottom.x = g_selection.x + g_selection.width;
		g_selection_bottom.y = g_selection.y + g_selection.height;
    }

    switch( event )
    {
    case CV_EVENT_LBUTTONDOWN:
        g_origin = cvPoint(x,y);
        g_selection = cvRect(x,y,0,0);
        g_select_object = 1;
        break;
    case CV_EVENT_LBUTTONUP:
        g_select_object = 0;
        if( g_selection.width > 0 && g_selection.height > 0 )
			g_range_selected = 1;
        break;
    }
}



void removeNonColors(IplImage *src, IplImage* dst, int r, int g, int b)
{
	int ii;
	int delta = dst->widthStep - dst->width*3 - 3;
	char *src_pic = src->imageData;
	char *dst_pic = dst->imageData;

	for (ii = 0; ii < dst->imageSize; ii += delta)
	{
		for (int x = 0; x < dst->width / 3; x +=3, ii += 3)
		{
			float r, g, b;
			int ir, ig, ib;
			
			r = src->imageData[ii];
			g = src->imageData[ii + 1];
			b = src->imageData[ii + 2];
			float total = r + g + b;
			float r_ratio = r / total;
			float g_ratio = r / total;
			float b_ratio = r / total;
			if (r > 10 && r > 0.4 && g < 0.31 && b < 0.36)
			{
				ir = ig = ib = 255;
			}
			else
			{
				ir = ig = ib = 0;
			}
			dst->imageData[ii] = ir;
			dst->imageData[ii + 1] = ig;
			dst->imageData[ii + 2] = ib;
		}
	}
}

// various tracking parameters (in seconds)
const double MHI_DURATION = 1;
const double MAX_TIME_DELTA = 0.5;
const double MIN_TIME_DELTA = 0.05;
// number of cyclic frame buffer used for motion detection
// (should, probably, depend on FPS)
const int N = 4;

// ring image buffer
IplImage **buf = 0;
int last = 0;

// temporary images
IplImage *mhi = 0; // MHI
IplImage *orient = 0; // orientation
IplImage *mask = 0; // valid orientation mask
IplImage *segmask = 0; // motion segmentation map
CvMemStorage* storage = 0; // temporary storage

// parameters:
//  img - input video frame
//  dst - resultant motion picture
//  args - optional parameters
void update_mhi( IplImage* img, IplImage* tmp, IplImage* dst, int diff_threshold )
{
    double timestamp = (double)clock()/CLOCKS_PER_SEC; // get current time in seconds
    CvSize size = cvSize(img->width,img->height); // get current frame size
    int i, idx1 = last, idx2;
    IplImage* silh;
    CvSeq* seq;
    CvRect comp_rect;
    double count;
    double angle;
    double magnitude;          
    CvScalar color;

    // allocate images at the beginning or
    // reallocate them if the frame size is changed
    if( !mhi || mhi->width != size.width || mhi->height != size.height ) {
        if( buf == 0 ) {
            buf = (IplImage**)malloc(N*sizeof(buf[0]));
            memset( buf, 0, N*sizeof(buf[0]));
        }
        
        for( i = 0; i < N; i++ ) {
            cvReleaseImage( &buf[i] );
            buf[i] = cvCreateImage( size, IPL_DEPTH_8U, 1 );
            cvZero( buf[i] );
        }
        cvReleaseImage( &mhi );
        cvReleaseImage( &orient );
        cvReleaseImage( &segmask );
        cvReleaseImage( &mask );
        
        mhi = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        cvZero( mhi ); // clear MHI at the beginning
        orient = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        segmask = cvCreateImage( size, IPL_DEPTH_32F, 1 );
        mask = cvCreateImage( size, IPL_DEPTH_8U, 1 );
    }

    cvCvtColor( img, buf[last], CV_BGR2GRAY ); // convert frame to grayscale
//	cvCopy(img, buf[last]);
    idx2 = (last + 1) % N; // index of (last - (N-1))th frame
    last = idx2;

    silh = buf[idx2];

	cvAbsDiff( buf[idx1], buf[idx2], silh ); // get difference between frames
	//cvCopy(silh, dst);
	//cvThreshold( silh, silh, diff_threshold, 1, CV_THRESH_TOZERO); // and threshold it
	
	//removeNonColors(silh, dst, 255, 0, 0);
	//return;
	cvThreshold(silh, silh, diff_threshold, 1, CV_THRESH_BINARY); // and threshold it

    cvUpdateMotionHistory( silh, mhi, timestamp, MHI_DURATION ); // update MHI

    // convert MHI to blue 8u image
    cvCvtScale( mhi, mask, 255./MHI_DURATION,
                (MHI_DURATION - timestamp)*255./MHI_DURATION );
    cvZero( dst );
    cvCvtPlaneToPix( mask, 0, 0, 0, dst );

    // calculate motion gradient orientation and valid orientation mask
    cvCalcMotionGradient( mhi, mask, orient, MAX_TIME_DELTA, MIN_TIME_DELTA, 3 );
    
    if( !storage )
        storage = cvCreateMemStorage(0);
    else
        cvClearMemStorage(storage);
    
    // segment motion: get sequence of motion components
    // segmask is marked motion components map. It is not used further
    seq = cvSegmentMotion( mhi, segmask, storage, timestamp, MAX_TIME_DELTA * 0.25);

    // iterate through the motion components,
    // One more iteration (i == -1) corresponds to the whole image (global motion)
    for( i = 0; i < seq->total; i++ ) {

        if( i < 0 ) { // case of the whole image
            comp_rect = cvRect( 0, 0, size.width, size.height );
            color = CV_RGB(255,255,255);
            magnitude = 100;
        }
        else { // i-th motion component
			CvConnectedComp* con = (CvConnectedComp*)cvGetSeqElem( seq, i );
            comp_rect = con->rect;
            if( comp_rect.width + comp_rect.height > 200 ) // reject very large components
                continue;
            color = CV_RGB(255,0,0);
            magnitude = 30;
        }

        // select component ROI
        cvSetImageROI( silh, comp_rect );
        cvSetImageROI( mhi, comp_rect );
        cvSetImageROI( orient, comp_rect );
        cvSetImageROI( mask, comp_rect );

        // calculate orientation
        angle = cvCalcGlobalOrientation( orient, mask, mhi, timestamp, MHI_DURATION);
        angle = 360.0 - angle;  // adjust for images with top-left origin

        count = cvNorm( silh, 0, CV_L1, 0 ); // calculate number of points within silhouette ROI

/*		if (count < 10)
		{
			CvPoint p1, p2;
			p1.x = comp_rect.x;
			p1.y = comp_rect.y;
			p2.x = comp_rect.width + comp_rect.x;
			p2.y = comp_rect.height + comp_rect.y;
			cvRectangle(dst, p1, p2, CV_RGB(255,255,255), -1);
		}*/

        cvResetImageROI( mhi );
        cvResetImageROI( orient );
        cvResetImageROI( mask );
        cvResetImageROI( silh );
	}
}

struct Point
{
	double x;
	double y;
	Point() : x(-1.0), y(-1.0) {}
	Point(double _x, double _y) : x(_x), y(_y) {}
};

#define NUM_POINTS 4UL // must be a power of 2

void CalcAvgPos(const Point *Points, int &mouse_pos_x, int &mouse_pos_y)
{
	int num_points = 0;
	double avg_x = 0.0, avg_y = 0.0;
	for (int ii = 0; ii < NUM_POINTS; ++ii)
	{
		if (Points[ii].x < 0 || Points[ii].y < 0)
			continue;
		avg_x += Points[ii].x;
		avg_y += Points[ii].y;
		++num_points;
	}

	avg_x /= num_points;
	avg_y /= num_points;

	mouse_pos_x = (int)floor(avg_x + 0.5);
	mouse_pos_y = (int)floor(avg_y + 0.5);
}


int main(int argc, char** argv)
{
	int resx = 0, resy = 0, last_x = 0, last_y = 0, mouse_pos_x = 1, mouse_pos_y = 1;
	resx = GetSystemMetrics(SM_CXSCREEN);
	resy = GetSystemMetrics(SM_CYSCREEN);

	Point Points[NUM_POINTS];
	unsigned int cur_point = 0;
//    IplImage* motion = 0;
//    CvCapture* capture = 0;
    
/*    if( argc == 1 || (argc == 2 && strlen(argv[1]) == 1 && isdigit(argv[1][0])))
        capture = cvCaptureFromCAM( argc == 2 ? argv[1][0] - '0' : 0 );
    else if( argc == 2 )
        capture = cvCaptureFromFile( argv[1] );*/

	if (!videoInput::listDevices()) return 1;

	videoInput* VI = new videoInput();

	VI->setupDevice(0);

    //if( capture )
    {
        //cvNamedWindow( "Motion", 1 );
        cvNamedWindow( "Orig", 1 );

		int width = VI->getWidth(0);
		int height = VI->getHeight(0);

		cout << width << ", " << height << endl;

		IplImage* newImage = cvCreateImage(cvSize(width, height), 8 ,3);
		IplImage* tmp = cvCreateImage(cvSize(width, height), 8 ,1);
        int ii = 0;
        g_image = cvCloneImage(newImage);

		cvSetMouseCallback( "Orig", on_mouse, 0 );
		int last_laser_on_duration = 0;
		int laser_on_duration = 0;
		DWORD tick_count = GetTickCount();
		DWORD last_tick_count;
		DWORD tick_delta; // milliseconds

		int time_laser_on = 0;
		int time_laser_on_last = 0;
		bool laser_state = false;
		bool last_laser_state = false;
		bool click_state = false;
		bool click_right_button = false;
		int last_x_good_position = 0;
		int last_y_good_position = 0;

		for(;;++ii)
        {
			last_tick_count = tick_count;
			tick_count = GetTickCount();
			if (tick_count > last_tick_count)
			{
				tick_delta = tick_count - last_tick_count;
			}
			else
			{
				tick_delta = tick_count + (UINT_MAX - last_tick_count);
			}

//			std::cout << ii << std::endl;
			//if (VI->isFrameNew(0));
//            if( !cvGrabFrame( capture ))
//                break;
//            image = cvRetrieveFrame( capture );
			VI->getPixels(0, (unsigned char*)g_image->imageData, false, true);
			//cvSmooth(image, newImage);
			
			//image = CvFilter(;
/*            if( g_image )
            {
                if( !motion )
                {
                    motion = cvCreateImage( cvSize(g_image->width, g_image->height), 8, 3 );
                    cvZero( motion );
                    motion->origin = g_image->origin;
                }
            }*/

//            update_mhi(g_image, tmp, motion, 7);

			double pos_x, pos_y, pic_pos_x, pic_pos_y;

			bool is_laser_on = GetLaserPos(g_image, pic_pos_x, pic_pos_y);

			last_laser_state = laser_state;
			laser_state = is_laser_on;

			if(!g_select_object && g_range_selected && is_laser_on)
			{
				time_laser_on += tick_delta;

				if (pic_pos_x < g_selection_bottom.x && pic_pos_y < g_selection_bottom.y &&
					pic_pos_x >= g_selection.x && pic_pos_y >= g_selection.y)
				{
					pos_x = (pic_pos_x - g_selection.x) / (double) g_selection.width;
					pos_y = (pic_pos_y - g_selection.y) / (double) g_selection.height;

					//mouse_pos_x = (int);
					//mouse_pos_y = (int)floor((pos_y * resy) + 0.5);
					Points[cur_point].x = (pos_x * 65535 * 1.1);
					Points[cur_point].y = (pos_y * 65535 * 1.1);
					++cur_point;
					cur_point &= (NUM_POINTS - 1);
					CalcAvgPos(Points, mouse_pos_x, mouse_pos_y);
				}
			}
			else
			{
//				cout << "click state: " << click_state << endl;
//				cout << "time_laser_on: " << time_laser_on << endl;
				if (laser_state == last_laser_state)
				{
					time_laser_on_last = time_laser_on;
					time_laser_on = 0;
					if (!laser_state && time_laser_on_last > 50 && time_laser_on_last < 250)
					{
						INPUT mi;
						mi.type = INPUT_MOUSE;
						mi.mi.dx = last_x_good_position;
						mi.mi.dy = last_y_good_position;
						mi.mi.mouseData = 0;
						mi.mi.dwFlags = 0;
						mi.mi.time = 0;
						mi.mi.dwExtraInfo = 0;
						mi.mi.dwFlags = MOUSEEVENTF_MOVE | MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_VIRTUALDESK;
						SendInput(1, &mi, sizeof(mi));

						if (!click_state && mouse_pos_x > 57343 && mouse_pos_y > 57343)
						{
							click_right_button = !click_right_button;
						}
						else
						{
							click_state = !click_state;
							cout << "click state: " << click_state << endl;
							INPUT mi;
							mi.type = INPUT_MOUSE;
							mi.mi.dx = 0;
							mi.mi.dy = 0;
							mi.mi.mouseData = 0;
							mi.mi.dwFlags = 0;
							mi.mi.time = 0;
							mi.mi.dwExtraInfo = 0;
							if (click_right_button)
							{
								mi.mi.dwFlags = MOUSEEVENTF_RIGHTDOWN;
								SendInput(1, &mi, sizeof(mi));
								mi.mi.dwFlags = MOUSEEVENTF_RIGHTUP;
								SendInput(1, &mi, sizeof(mi));
								click_state = false;
								click_right_button = false;
							}
							else
							{
								mi.mi.dwFlags = (click_state ? MOUSEEVENTF_LEFTDOWN : MOUSEEVENTF_LEFTUP);
								SendInput(1, &mi, sizeof(mi));
							}
						}
					}
					else
					{
						if (!laser_state)
						{
							last_x_good_position = mouse_pos_x;
							last_y_good_position = mouse_pos_y;
						}
					}
				}
			}

			if(g_select_object && g_selection.width > 0 && g_selection.height > 0 )
			{
				cvSetImageROI( g_image, g_selection );
				cvXorS( g_image, cvScalarAll(255), g_image, 0 );
				cvResetImageROI( g_image );
			}


			if(g_range_selected && g_selection.width > 0 && g_selection.height > 0
				&& (last_x != mouse_pos_x || last_y != mouse_pos_y))
			{
//				SetCursorPos(mouse_pos_x, mouse_pos_y);
				INPUT mi;
				mi.type = INPUT_MOUSE;
				mi.mi.dx = mouse_pos_x;
				mi.mi.dy = mouse_pos_y;
				mi.mi.mouseData = 0;
				mi.mi.dwFlags = 0;
				mi.mi.time = 0;
				mi.mi.dwExtraInfo = 0;
				mi.mi.dwFlags = MOUSEEVENTF_MOVE | MOUSEEVENTF_ABSOLUTE | MOUSEEVENTF_VIRTUALDESK;
				SendInput(1, &mi, sizeof(mi));

				last_x = mouse_pos_x;
				last_y = mouse_pos_y;
			}


            cvShowImage( "Orig", g_image );
//            cvShowImage( "Motion", motion );

            if( cvWaitKey(10) >= 0 )
                break;
        }
        //cvReleaseCapture( &capture );
        cvDestroyWindow( "Orig" );
 //       cvDestroyWindow( "Motion" );
    }

	delete VI;
    return 0;
}
